import math

import matplotlib
import matplotlib.cm as plt_cm
import numpy as np

from pydrake.common.value import Value
from pydrake.geometry import (
    Box,
    Convex,
    Cylinder,
    HalfSpace,
    Mesh,
    QueryObject,
    Rgba,
    Role,
    Sphere,
    optimization,
)
from pydrake.math import RigidTransform
from pydrake.systems.pyplot_visualizer import PyPlotVisualizer


class PlanarSceneGraphVisualizer(PyPlotVisualizer):
    """
    Given a SceneGraph and a view plane, provides a view of the robot by
    projecting all geometry onto the view plane.

    This is intended to be used for robots that operate in the plane, but
    should render any robot approximately correctly. It has the following
    caveats:
    - z-ordering of objects is done based on the object centroid, which
    is not perfect for non-planar scenes.
    - Object geometry is projected onto the view plane, then a chull is taken,
    and finally that chull is drawn as a patch. Nonconvex geometry will thus be
    drawn incorrectly, and geometry with many vertices will slow down the
    visualizer.

    Specifics on view setup:

    T_VW specifies the 3x4 view projection matrix. For planar orthographic
    projection, use:
    [ <x axis select> x_axis_shift
      <y axis select> y_axis_shift
       0, 0, 0, 1]  % homogenizer

    e.g.

    [ 1 0 0 0.5
      0 1 0 0
      0 0 0 1].

    would give a top-down view (i.e squashing the z axis), and would shift
    things in the x axis positively by 0.5.

    T_VW can be any valid view projection matrix. If the bottom row is
    [0, 0, 0, 1], the view projection will be an orthographic projection.

    xlim and ylim don't technically provide extra functionality, but it's
    easier to keep handling scaling with xlim, ylim, and view plane selection
    and *maybe* offsetting with the projection matrix.
    """

    def __init__(
        self,
        scene_graph,
        draw_period=None,
        T_VW=np.array(
            [
                [1.0, 0.0, 0.0, 0.0],  # BR
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        ),
        xlim=[-1.0, 1],
        ylim=[-1, 1],
        facecolor=[1, 1, 1],
        use_random_colors=False,
        substitute_collocated_mesh_files=True,
        ax=None,
        show=None,
    ):
        """
        Args:
            scene_graph: A SceneGraph object.
            draw_period: The rate at which this class publishes to the
                visualizer.  When None, a suitable default will be used.
            T_VW: The view projection matrix from world to view coordinates.
            xlim: View limit into the scene.
            ylim: View limit into the scene.
            facecolor: Passed through to figure() and sets background color.
                Both color name strings and RGB triplets are allowed. Defaults
                to white.
            use_random_colors: If set to True, will render each body with a
                different color. (Multiple visual elements on the same body
                will be the same color.)
            substitute_collocated_mesh_files: If True, then a mesh file
                specified with an unsupported filename extension may be
                replaced by a file of the same base name in the same
                directory, but with a supported filename extension.
            ax: If supplied, the visualizer will draw onto those axes instead
                of creating a new set of axes. The visualizer will still change
                the view range and figure size of those axes.
            show: Opens a window during initialization / publish iff True.
                Default is None, which implies show=True unless
                matplotlib.get_backend() is 'template'.
        """
        default_size = matplotlib.rcParams["figure.figsize"]
        scalefactor = (ylim[1] - ylim[0]) / (xlim[1] - xlim[0])
        figsize = (default_size[0], default_size[0] * scalefactor)

        PyPlotVisualizer.__init__(
            self,
            facecolor=facecolor,
            figsize=figsize,
            ax=ax,
            draw_period=draw_period,
            show=show,
        )
        self.set_name("planar_scenegraph_visualizer")

        self._scene_graph = scene_graph
        self._T_VW = T_VW

        self._geometry_query_input_port = self.DeclareAbstractInputPort(
            "geometry_query", Value(QueryObject())
        )

        self.ax.axis("equal")
        self.ax.axis("off")

        # Achieve the desired view limits.
        self.ax.set_xlim(xlim)
        self.ax.set_ylim(ylim)
        default_size = self.fig.get_size_inches()
        self.fig.set_size_inches(figsize[0], figsize[1])

        # Populate body patches.
        self._build_body_patches(
            use_random_colors,
            substitute_collocated_mesh_files,
            scene_graph.model_inspector(),
        )

        # Populate the body fill list -- which requires doing most of a draw
        # pass, but with an ax.fill() command to initialize the draw patches.
        # After initialization, we can then use in-place replacement of vertex
        # positions. The body fill list stores the ax patch objects in the
        # order they were spawned (i.e. by body, and then by order of view_
        # patches). Drawing the tree should update them by iterating over
        # bodies and patches in the same order.
        self._body_fill_dict = {}
        X_WB_initial = RigidTransform.Identity()
        for full_name in self._patch_Blist.keys():
            patch_Wlist, view_colors = self._get_view_patches(
                full_name, X_WB_initial
            )
            self._body_fill_dict[full_name] = []
            for patch_W, color in zip(patch_Wlist, view_colors):
                # Project the full patch the first time, to initialize a vertex
                # list with enough space for any possible convex hull of this
                # vertex set.
                patch_V = self._project_patch(patch_W)
                body_fill = self.ax.fill(
                    patch_V[0, :],
                    patch_V[1, :],
                    zorder=0,
                    edgecolor="k",
                    facecolor=color,
                    closed=True,
                )[0]
                self._body_fill_dict[full_name].append(body_fill)
                # Then update the vertices for a more accurate initial draw.
                self._update_body_fill_verts(body_fill, patch_V)

    def get_geometry_query_input_port(self):
        return self._geometry_query_input_port

    @staticmethod
    def frame_name(frame_id, inspector):
        """Produces a visualizer name for a frame."""
        # The frame names have been selected to match the names produced by
        # DrakeVisualizer. That is *not* a requirement, and this visualizer can
        # choose a different naming scheme.
        if frame_id == inspector.world_frame_id():
            return "world"
        return "{}::{}".format(
            inspector.GetOwningSourceName(frame_id), inspector.GetName(frame_id)
        )

    def _build_body_patches(
        self, use_random_colors, substitute_collocated_mesh_files, inspector
    ):
        """
        Generates body patches. self._patch_Blist stores a list of patches for
        each body (starting at body id 1). A body patch is a list of all 3D
        vertices of a piece of visual geometry.
        """
        self._patch_Blist = {}
        self._patch_Blist_colors = {}

        for frame_id in inspector.GetAllFrameIds():
            count = inspector.NumGeometriesForFrameWithRole(
                frame_id, Role.kIllustration
            )
            if count == 0:
                continue

            frame_name = self.frame_name(frame_id, inspector)

            this_body_patches = []
            this_body_colors = []

            for g_id in inspector.GetGeometries(frame_id, Role.kIllustration):
                X_BG = inspector.GetPoseInFrame(g_id)
                shape = inspector.GetShape(g_id)

                if isinstance(shape, Box):
                    # Draw a bounding box.
                    half_width = shape.width() / 2.0
                    half_depth = shape.depth() / 2.0
                    half_height = shape.height() / 2.0
                    patch_G = np.vstack(
                        (
                            half_width * np.array([-1, -1, 1, 1, -1, -1, 1, 1]),
                            half_depth * np.array([-1, 1, -1, 1, -1, 1, -1, 1]),
                            half_height * np.array([-1, -1, -1, -1, 1, 1, 1, 1]),  # noqa
                        )
                    )  # fmt: skip

                elif isinstance(shape, Sphere):
                    # Sphere is the only shape that allows a zero-measure. A
                    # zero-radius sphere is a point, and we skip it.
                    if shape.radius() == 0:
                        continue

                    lati, longi = np.meshgrid(
                        np.arange(0.0, 2.0 * math.pi, 0.5),
                        np.arange(0.0, 2.0 * math.pi, 0.5),
                    )
                    lati = lati.ravel()
                    longi = longi.ravel()
                    patch_G = np.vstack(
                        [
                            np.sin(lati) * np.cos(longi),
                            np.sin(lati) * np.sin(longi),
                            np.cos(lati),
                        ]
                    )
                    patch_G *= shape.radius()

                elif isinstance(shape, Cylinder):
                    radius = shape.radius()
                    length = shape.length()

                    # In the lcm geometry, cylinders are along +z
                    # https://github.com/RobotLocomotion/drake/blob/last_sha_with_original_matlab/drake/matlab/systems/plants/RigidBodyCylinder.m
                    # Two circles: one at bottom, one at top.
                    sample_cos_sin = [
                        (math.cos(angle), math.sin(angle))
                        for angle in np.arange(0.0, 2.0 * math.pi, 0.25)
                    ]
                    patch_G = np.hstack(
                        [
                            np.array(
                                [
                                    [radius * cos, radius * sin, -length / 2.0],
                                    [radius * cos, radius * sin, length / 2.0],
                                ]
                            ).T
                            for cos, sin in sample_cos_sin
                        ]
                    )

                elif isinstance(shape, (Mesh, Convex)):
                    source = shape.source()
                    search_for_alternate = (
                        source.is_path() and substitute_collocated_mesh_files
                    )
                    convex_hull = None
                    try:
                        # For both shape types, we replace it with its convex
                        # hull.
                        convex_hull = shape.GetConvexHull()
                    except RuntimeError as shape_error:
                        known_suffixes = [".obj", ".vtk", ".gltf"]

                        if source.extension() in known_suffixes:
                            # The file was already of a known extension;
                            # failure offers no recourse.
                            raise shape_error

                        if search_for_alternate:
                            # For a path to an unknown file type, we can look
                            # for a known alternative.
                            for suffix in known_suffixes:
                                alt_path = source.path().with_suffix(suffix)
                                if alt_path.exists():
                                    convex = Convex(alt_path, shape.scale3())
                                    convex_hull = convex.GetConvexHull()
                                    break

                    if convex_hull is None:
                        # If we're here, we know:
                        #   shape.source().extension() is not supported.
                        #   We didn't/couldn't find a supported alternative.
                        suffix = (
                            " No supported alternative could be found."
                            if search_for_alternate
                            else ""
                        )
                        raise RuntimeError(
                            f"The {type(shape).__name__} instance with mesh "
                            f"data '{source.description()}' has an "
                            f"unsupported extension ('{source.extension()}')."
                            f"{suffix}"
                        )
                    patch_G = np.empty((3, convex_hull.num_vertices()))
                    for i in range(convex_hull.num_vertices()):
                        patch_G[:, i] = convex_hull.vertex(i)

                elif isinstance(shape, HalfSpace):
                    # For a half space, we'll simply create a large box with
                    # the top face at z = 0, the bottom face at z = -1 and the
                    # far corners at +/- 50 in the x- and y- directions.
                    x = 50
                    y = 50
                    z = -1
                    patch_G = np.vstack(
                        (
                            x * np.array([-1, -1, 1, 1, -1, -1, 1, 1]),
                            y * np.array([-1, 1, -1, 1, -1, 1, -1, 1]),
                            z * np.array([-1, -1, -1, -1, 0, 0, 0, 0]),
                        )
                    )

                # TODO(SeanCurtis-TRI): Provide support for capsule and
                # ellipsoid.
                else:
                    print(
                        "UNSUPPORTED GEOMETRY TYPE {} IGNORED".format(
                            type(shape)
                        )
                    )
                    continue

                # Compute pose in body.
                patch_B = X_BG @ patch_G

                # Close path if not closed.
                if (patch_B[:, -1] != patch_B[:, 0]).any():
                    patch_B = np.hstack((patch_B, patch_B[:, 0][np.newaxis].T))

                this_body_patches.append(patch_B)
                if not use_random_colors:
                    # If we need to use random colors, we apply them after the
                    # fact. See below.
                    props = inspector.GetIllustrationProperties(g_id)
                    assert props is not None
                    rgba = props.GetPropertyOrDefault(
                        "phong", "diffuse", Rgba(0.9, 0.9, 0.9, 1.0)
                    )
                    color = np.array((rgba.r(), rgba.g(), rgba.b(), rgba.a()))
                    this_body_colors.append(color)

            self._patch_Blist[frame_name] = this_body_patches
            self._patch_Blist_colors[frame_name] = this_body_colors

        # Spawn a random color generator. Each body will be given a unique
        # color when using this random generator, with each visual element of
        # the body colored the same.
        if use_random_colors:
            color = iter(
                plt_cm.rainbow(np.linspace(0, 1, len(self._patch_Blist_colors)))
            )
            for name in self._patch_Blist_colors.keys():
                this_color = next(color)
                patch_count = len(self._patch_Blist[name])
                self._patch_Blist_colors[name] = [this_color] * patch_count

    def _get_view_patches(self, full_name, X_WB):
        """
        Pulls out the view patch verts for the given body index after applying
        the appropriate transform, X_WB. X_WB needs to be a RigidTransform.
        """
        patch_Wlist = []
        for patch_B in self._patch_Blist[full_name]:
            patch_W = X_WB @ patch_B
            # Add homogeneous row.
            patch_W = np.vstack((patch_W, np.ones((1, patch_W.shape[1]))))
            patch_Wlist.append(patch_W)

        colors = self._patch_Blist_colors[full_name]
        return (patch_Wlist, colors)

    def _project_patch(self, patch_W):
        """
        Project the object vertices from 3d in world frame W to 2d in view
        frame V.
        """
        patch_V = self._T_VW @ patch_W
        # Applies normalization in the perspective transformation
        # to make each projected point have z = 1. If the bottom row
        # of T_VW is [0, 0, 0, 1], this will result in an
        # orthographic projection.
        patch_V[0, :] /= patch_V[2, :]
        patch_V[1, :] /= patch_V[2, :]
        # Cut patch_V down to 2xN.
        patch_V = patch_V[:2, :]
        return patch_V

    def _update_body_fill_verts(self, body_fill, patch_V):
        """
        Takes a convex hull if necessary and uses in-place replacement of
        vertices to update the fill.
        """

        # Take a convex hull to get an accurate shape for drawing, with verts
        # coming out in ccw order.
        if patch_V.shape[1] > 3:
            vpoly = optimization.VPolytope(patch_V)
            patch_V = vpoly.GetMinimalRepresentation().vertices()

        # Update the verts, padding out to the appropriate full # of verts by
        # replicating the final vertex.
        n_verts = body_fill.get_path().vertices.shape[0]
        patch_V = np.pad(
            patch_V, ((0, 0), (0, n_verts - patch_V.shape[1])), mode="edge"
        )
        body_fill.get_path().vertices[:, :] = patch_V.T

    def draw(self, context):
        """Overrides base with the implementation."""
        query_object = self._geometry_query_input_port.Eval(context)
        inspector = query_object.inspector()

        view_dir = np.cross(self._T_VW[0, :3], self._T_VW[1, :3])
        for frame_id in inspector.GetAllFrameIds():
            frame_name = self.frame_name(frame_id, inspector)
            if frame_name not in self._patch_Blist:
                continue
            X_WB = query_object.GetPoseInWorld(frame_id)

            patch_Wlist, _ = self._get_view_patches(frame_name, X_WB)
            for i, patch_W in enumerate(patch_Wlist):
                # Project the object vertices from 3d in world frame W to 2d in
                # view frame V (keeps homogeneous portion, removing it later).
                patch_V = self._project_patch(patch_W)
                body_fill = self._body_fill_dict[frame_name][i]
                # Use the latest vertices to update the body_fill.
                self._update_body_fill_verts(body_fill, patch_V)
                body_fill.zorder = X_WB.translation() @ view_dir
        self.ax.set_title("t = {:.1f}".format(context.get_time()))


def ConnectPlanarSceneGraphVisualizer(
    builder, scene_graph, output_port=None, **kwargs
):
    """Creates an instance of PlanarSceneGraphVisualizer, adds it to the
    diagram, and wires the scene_graph pose bundle output port to the input
    port of the visualizer.  Provides an interface comparable to
    DrakeVisualizer.AddToBuilder.

    Args:
        builder: The diagram builder used to construct the Diagram.
        scene_graph: The SceneGraph in builder containing the geometry to be
            visualized.
        output_port: (optional) If not None, then output_port will be connected
            to the visualizer input port instead of the scene_graph.
            get_query_output_port().  This is required, for instance,
            when the SceneGraph is inside a Diagram, and we must connect the
            exposed port to the visualizer instead of the original SceneGraph
            port.

        Additional kwargs are passed through to the PlanarSceneGraphVisualizer
        constructor.

    Returns:
        The newly created PlanarSceneGraphVisualizer object.
    """
    visualizer = builder.AddSystem(
        PlanarSceneGraphVisualizer(scene_graph, **kwargs)
    )

    if output_port is None:
        output_port = scene_graph.get_query_output_port()

    builder.Connect(output_port, visualizer.get_geometry_query_input_port())
    return visualizer



"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_panda_command(object):
    """
    Commands the desired state of a Franka Panda arm.
    
    The Franka Control Interface provides a number of possible modes to control
    the arm, as described in
    
    https://frankaemika.github.io/docs/libfranka.html#realtime-commands
    
    The expected control mode is given by the control_mode_expected, below.
    
    The num_joint_{position,velocity,torque} size must not have multiple
    different non-zero values.  Each one must be set to either the actual
    number of joints (e.g., 7) or else zero.
    """

    __slots__ = ["utime", "num_joint_position", "joint_position", "num_joint_velocity", "joint_velocity", "num_joint_torque", "joint_torque", "control_mode_expected"]

    __typenames__ = ["int64_t", "int32_t", "double", "int32_t", "double", "int32_t", "double", "int8_t"]

    __dimensions__ = [None, None, ["num_joint_position"], None, ["num_joint_velocity"], None, ["num_joint_torque"], None]

    def __init__(self):
        self.utime = 0
        """
        The timestamp in microseconds.
        LCM Type: int64_t
        """

        self.num_joint_position = 0
        """
        The commanded joint positions in radians.
        
        These values must be provided if the CONTROL_MODE_POSITION bit is set
        within control_mode_expected.
        LCM Type: int32_t
        """

        self.joint_position = []
        """ LCM Type: double[num_joint_position] """
        self.num_joint_velocity = 0
        """
        The commanded joint velocities in radians per second.
        
        These values must be provided if the CONTROL_MODE_VELOCITY bit is set
        within control_mode_expected.
        LCM Type: int32_t
        """

        self.joint_velocity = []
        """ LCM Type: double[num_joint_velocity] """
        self.num_joint_torque = 0
        """
        The commanded joint torques.
        
        These values must be provided if the CONTROL_MODE_TORQUE bit is set
        within control_mode_expected.
        LCM Type: int32_t
        """

        self.joint_torque = []
        """ LCM Type: double[num_joint_torque] """
        self.control_mode_expected = 0
        """
        Describes how the controller expects the driver to be configured.  See
        the values for control_mode in lcmt_panda_status.
        LCM Type: int8_t
        """


    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_panda_command._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">qi", self.utime, self.num_joint_position))
        buf.write(struct.pack('>%dd' % self.num_joint_position, *self.joint_position[:self.num_joint_position]))
        buf.write(struct.pack(">i", self.num_joint_velocity))
        buf.write(struct.pack('>%dd' % self.num_joint_velocity, *self.joint_velocity[:self.num_joint_velocity]))
        buf.write(struct.pack(">i", self.num_joint_torque))
        buf.write(struct.pack('>%dd' % self.num_joint_torque, *self.joint_torque[:self.num_joint_torque]))
        buf.write(struct.pack(">b", self.control_mode_expected))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_panda_command._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_panda_command._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_panda_command()
        self.utime, self.num_joint_position = struct.unpack(">qi", buf.read(12))
        self.joint_position = struct.unpack('>%dd' % self.num_joint_position, buf.read(self.num_joint_position * 8))
        self.num_joint_velocity = struct.unpack(">i", buf.read(4))[0]
        self.joint_velocity = struct.unpack('>%dd' % self.num_joint_velocity, buf.read(self.num_joint_velocity * 8))
        self.num_joint_torque = struct.unpack(">i", buf.read(4))[0]
        self.joint_torque = struct.unpack('>%dd' % self.num_joint_torque, buf.read(self.num_joint_torque * 8))
        self.control_mode_expected = struct.unpack(">b", buf.read(1))[0]
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_panda_command in parents: return 0
        tmphash = (0x7346fbbb8026ace8) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_panda_command._packed_fingerprint is None:
            lcmt_panda_command._packed_fingerprint = struct.pack(">Q", lcmt_panda_command._get_hash_recursive([]))
        return lcmt_panda_command._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_panda_command._get_packed_fingerprint())[0]




import collections.abc
import enum
import nanobind
import pydrake.autodiffutils
import pydrake.solvers
import pydrake.symbolic
import pydrake.systems.framework
import pydrake.trajectories
from _typeshed import Incomplete
from pydrake.common.cpp_template import IntegratorBase_ as IntegratorBase_, RungeKutta2Integrator_ as RungeKutta2Integrator_, RungeKutta3Integrator_ as RungeKutta3Integrator_, Simulator_ as Simulator_
from typing import Any, Callable, ClassVar, overload

ApplySimulatorConfig: nanobind.nb_func
BatchEvalTimeDerivatives: nanobind.nb_func
BatchEvalUniquePeriodicDiscreteUpdate: nanobind.nb_func
DiscreteTimeApproximation: nanobind.nb_func
ExtractSimulatorConfig: nanobind.nb_func
GetIntegrationSchemes: nanobind.nb_func
MonteCarloSimulation: nanobind.nb_func
PrintSimulatorStatistics: nanobind.nb_func
RandomSimulation: nanobind.nb_func
RegionOfAttraction: nanobind.nb_func
ResetIntegratorFromFlags: nanobind.nb_func

class InitializeParams:
    suppress_initialization_events: bool
    def __init__(self, **kwargs) -> None: ...
    def __copy__(self) -> InitializeParams: ...
    def __deepcopy__(self, arg: dict) -> InitializeParams: ...
    @property
    def __fields__(self): ...

class IntegratorBase:
    def __init__(self, *args, **kwargs) -> None: ...
    def GetStatisticsSummary(self) -> list[tuple[str, int | float]]: ...
    @overload
    def Initialize(self) -> None: ...
    @overload
    def Initialize(self) -> Any: ...
    @overload
    def Initialize(self) -> Any: ...
    def Reset(self) -> None: ...
    def ResetStatistics(self) -> None: ...
    def StartDenseIntegration(self) -> None: ...
    def StopDenseIntegration(self) -> pydrake.trajectories.PiecewisePolynomial: ...
    def get_actual_initial_step_size_taken(self) -> float: ...
    def get_context(self) -> pydrake.systems.framework.Context: ...
    def get_dense_output(self) -> pydrake.trajectories.PiecewisePolynomial: ...
    def get_fixed_step_mode(self) -> bool: ...
    def get_initial_step_size_target(self) -> float: ...
    def get_largest_step_size_taken(self) -> float: ...
    def get_maximum_step_size(self) -> float: ...
    def get_mutable_context(self) -> pydrake.systems.framework.Context: ...
    def get_num_derivative_evaluations(self) -> int: ...
    def get_num_step_shrinkages_from_error_control(self) -> int: ...
    def get_num_step_shrinkages_from_substep_failures(self) -> int: ...
    def get_num_steps_taken(self) -> int: ...
    def get_num_substep_failures(self) -> int: ...
    def get_requested_minimum_step_size(self) -> float: ...
    def get_smallest_adapted_step_size_taken(self) -> float: ...
    def get_target_accuracy(self) -> float: ...
    def get_throw_on_minimum_step_size_violation(self) -> bool: ...
    def request_initial_step_size_target(self, step_size: float) -> None: ...
    def reset_context(self, context: pydrake.systems.framework.Context | None = ...) -> None: ...
    def set_fixed_step_mode(self, flag: bool) -> None: ...
    @overload
    def set_maximum_step_size(self, max_step_size: float) -> None: ...
    @overload
    def set_maximum_step_size(self) -> Any: ...
    def set_requested_minimum_step_size(self, min_step_size: float) -> None: ...
    def set_target_accuracy(self, accuracy: float) -> None: ...
    def set_throw_on_minimum_step_size_violation(self, throws: bool) -> None: ...

class IntegratorBase_𝓣AutoDiffXd𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def GetStatisticsSummary(self) -> list[tuple[str, int | float]]: ...
    @overload
    def Initialize(self) -> None: ...
    @overload
    def Initialize(self) -> Any: ...
    @overload
    def Initialize(self) -> Any: ...
    def Reset(self) -> None: ...
    def ResetStatistics(self) -> None: ...
    def StartDenseIntegration(self) -> None: ...
    def StopDenseIntegration(self) -> pydrake.trajectories.PiecewisePolynomial_TAutoDiffXdU: ...
    def get_actual_initial_step_size_taken(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_context(self) -> pydrake.systems.framework.Context_TAutoDiffXdU: ...
    def get_dense_output(self) -> pydrake.trajectories.PiecewisePolynomial_TAutoDiffXdU: ...
    def get_fixed_step_mode(self) -> bool: ...
    def get_initial_step_size_target(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_largest_step_size_taken(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_maximum_step_size(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_mutable_context(self) -> pydrake.systems.framework.Context_TAutoDiffXdU: ...
    def get_num_derivative_evaluations(self) -> int: ...
    def get_num_step_shrinkages_from_error_control(self) -> int: ...
    def get_num_step_shrinkages_from_substep_failures(self) -> int: ...
    def get_num_steps_taken(self) -> int: ...
    def get_num_substep_failures(self) -> int: ...
    def get_requested_minimum_step_size(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_smallest_adapted_step_size_taken(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_target_accuracy(self) -> float: ...
    def get_throw_on_minimum_step_size_violation(self) -> bool: ...
    def request_initial_step_size_target(self, step_size: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def reset_context(self, context: pydrake.systems.framework.Context_TAutoDiffXdU | None = ...) -> None: ...
    def set_fixed_step_mode(self, flag: bool) -> None: ...
    @overload
    def set_maximum_step_size(self, max_step_size: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    @overload
    def set_maximum_step_size(self) -> Any: ...
    def set_requested_minimum_step_size(self, min_step_size: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_target_accuracy(self, accuracy: float) -> None: ...
    def set_throw_on_minimum_step_size_violation(self, throws: bool) -> None: ...

class IntegratorBase_𝓣Expression𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def GetStatisticsSummary(self) -> list[tuple[str, int | float]]: ...
    @overload
    def Initialize(self) -> None: ...
    @overload
    def Initialize(self) -> Any: ...
    @overload
    def Initialize(self) -> Any: ...
    def Reset(self) -> None: ...
    def ResetStatistics(self) -> None: ...
    def StartDenseIntegration(self) -> None: ...
    def StopDenseIntegration(self) -> pydrake.trajectories.PiecewisePolynomial_TExpressionU: ...
    def get_actual_initial_step_size_taken(self) -> pydrake.symbolic.Expression: ...
    def get_context(self) -> pydrake.systems.framework.Context_TExpressionU: ...
    def get_dense_output(self) -> pydrake.trajectories.PiecewisePolynomial_TExpressionU: ...
    def get_fixed_step_mode(self) -> bool: ...
    def get_initial_step_size_target(self) -> pydrake.symbolic.Expression: ...
    def get_largest_step_size_taken(self) -> pydrake.symbolic.Expression: ...
    def get_maximum_step_size(self) -> pydrake.symbolic.Expression: ...
    def get_mutable_context(self) -> pydrake.systems.framework.Context_TExpressionU: ...
    def get_num_derivative_evaluations(self) -> int: ...
    def get_num_step_shrinkages_from_error_control(self) -> int: ...
    def get_num_step_shrinkages_from_substep_failures(self) -> int: ...
    def get_num_steps_taken(self) -> int: ...
    def get_num_substep_failures(self) -> int: ...
    def get_requested_minimum_step_size(self) -> pydrake.symbolic.Expression: ...
    def get_smallest_adapted_step_size_taken(self) -> pydrake.symbolic.Expression: ...
    def get_target_accuracy(self) -> float: ...
    def get_throw_on_minimum_step_size_violation(self) -> bool: ...
    def request_initial_step_size_target(self, step_size: pydrake.symbolic.Expression) -> None: ...
    def reset_context(self, context: pydrake.systems.framework.Context_TExpressionU | None = ...) -> None: ...
    def set_fixed_step_mode(self, flag: bool) -> None: ...
    @overload
    def set_maximum_step_size(self, max_step_size: pydrake.symbolic.Expression) -> None: ...
    @overload
    def set_maximum_step_size(self) -> Any: ...
    def set_requested_minimum_step_size(self, min_step_size: pydrake.symbolic.Expression) -> None: ...
    def set_target_accuracy(self, accuracy: float) -> None: ...
    def set_throw_on_minimum_step_size_violation(self, throws: bool) -> None: ...

class RandomSimulationResult:
    generator_snapshot: Incomplete
    output: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...

class RegionOfAttractionOptions:
    lyapunov_candidate: pydrake.symbolic.Expression
    solver_id: pydrake.solvers.SolverId | None
    solver_options: pydrake.solvers.SolverOptions | None
    state_variables: Incomplete
    use_implicit_dynamics: bool
    def __init__(self) -> None: ...
    def __copy__(self) -> RegionOfAttractionOptions: ...
    def __deepcopy__(self, arg: dict) -> RegionOfAttractionOptions: ...

class RungeKutta2Integrator(IntegratorBase):
    def __init__(self, system: pydrake.systems.framework.System, max_step_size: float, context: pydrake.systems.framework.Context | None = ...) -> None: ...

class RungeKutta2Integrator_𝓣AutoDiffXd𝓤(IntegratorBase_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TAutoDiffXdU, max_step_size: pydrake.autodiffutils.AutoDiffXd, context: pydrake.systems.framework.Context_TAutoDiffXdU | None = ...) -> None: ...

class RungeKutta2Integrator_𝓣Expression𝓤(IntegratorBase_TExpressionU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TExpressionU, max_step_size: pydrake.symbolic.Expression, context: pydrake.systems.framework.Context_TExpressionU | None = ...) -> None: ...

class RungeKutta3Integrator(IntegratorBase):
    def __init__(self, system: pydrake.systems.framework.System, context: pydrake.systems.framework.Context | None = ...) -> None: ...

class RungeKutta3Integrator_𝓣AutoDiffXd𝓤(IntegratorBase_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TAutoDiffXdU, context: pydrake.systems.framework.Context_TAutoDiffXdU | None = ...) -> None: ...

class Simulator:
    def __init__(self, system: pydrake.systems.framework.System, context: object | None = ...) -> None: ...
    def AdvancePendingEvents(self) -> SimulatorStatus: ...
    def AdvanceTo(self, boundary_time: float, interruptible: bool = ...) -> SimulatorStatus: ...
    def Initialize(self, params: InitializeParams = ...) -> SimulatorStatus: ...
    def ResetStatistics(self) -> None: ...
    def clear_monitor(self) -> None: ...
    def get_actual_realtime_rate(self) -> float: ...
    def get_context(self) -> pydrake.systems.framework.Context: ...
    def get_integrator(self) -> IntegratorBase: ...
    def get_monitor(self) -> collections.abc.Callable[[pydrake.systems.framework.Context], pydrake.systems.framework.EventStatus]: ...
    def get_mutable_context(self) -> pydrake.systems.framework.Context: ...
    def get_mutable_integrator(self) -> IntegratorBase: ...
    def get_num_discrete_updates(self) -> int: ...
    def get_num_publishes(self) -> int: ...
    def get_num_steps_taken(self) -> int: ...
    def get_num_unrestricted_updates(self) -> int: ...
    def get_system(self) -> pydrake.systems.framework.System: ...
    def get_target_realtime_rate(self) -> float: ...
    def has_context(self) -> bool: ...
    def reset_context(self, context: pydrake.systems.framework.Context | None) -> None: ...
    def set_monitor(self, monitor: collections.abc.Callable[[pydrake.systems.framework.Context], pydrake.systems.framework.EventStatus | None]) -> None: ...
    def set_target_realtime_rate(self, realtime_rate: float) -> None: ...

class SimulatorConfig:
    accuracy: Incomplete
    integration_scheme: Incomplete
    max_step_size: Incomplete
    start_time: float
    target_realtime_rate: Incomplete
    use_error_control: Incomplete
    def __init__(self, **kwargs) -> None: ...
    def __copy__(self) -> SimulatorConfig: ...
    def __deepcopy__(self, arg: dict) -> SimulatorConfig: ...
    @property
    def __fields__(self): ...

class SimulatorStatus:
    class ReturnReason(enum.Enum):
        __new__: ClassVar[Callable] = ...
        _generate_next_value_: ClassVar[Callable] = ...
        _hashable_values_: ClassVar[list] = ...
        _member_map_: ClassVar[dict] = ...
        _member_names_: ClassVar[list] = ...
        _member_type_: ClassVar[type[object]] = ...
        _unhashable_values_: ClassVar[list] = ...
        _unhashable_values_map_: ClassVar[dict] = ...
        _use_args_: ClassVar[bool] = ...
        _value2member_map_: ClassVar[dict] = ...
        _value_repr_: ClassVar[None] = ...
        kEventHandlerFailed: ClassVar[SimulatorStatus.ReturnReason] = ...
        kReachedBoundaryTime: ClassVar[SimulatorStatus.ReturnReason] = ...
        kReachedTerminationCondition: ClassVar[SimulatorStatus.ReturnReason] = ...
        __nb_enum__: ClassVar[PyCapsule] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def FormatMessage(self) -> str: ...
    def IsIdenticalStatus(self, other: SimulatorStatus) -> bool: ...
    def boundary_time(self) -> float: ...
    def message(self) -> str: ...
    def reason(self) -> SimulatorStatus.ReturnReason: ...
    def return_time(self) -> float: ...
    def succeeded(self) -> bool: ...
    def system(self) -> pydrake.systems.framework.SystemBase: ...

class Simulator_𝓣AutoDiffXd𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TAutoDiffXdU, context: object | None = ...) -> None: ...
    def AdvancePendingEvents(self) -> SimulatorStatus: ...
    def AdvanceTo(self, boundary_time: pydrake.autodiffutils.AutoDiffXd, interruptible: bool = ...) -> SimulatorStatus: ...
    def Initialize(self, params: InitializeParams = ...) -> SimulatorStatus: ...
    def ResetStatistics(self) -> None: ...
    def clear_monitor(self) -> None: ...
    def get_actual_realtime_rate(self) -> float: ...
    def get_context(self) -> pydrake.systems.framework.Context_TAutoDiffXdU: ...
    def get_integrator(self) -> IntegratorBase_TAutoDiffXdU: ...
    def get_monitor(self) -> collections.abc.Callable[[pydrake.systems.framework.Context_TAutoDiffXdU], pydrake.systems.framework.EventStatus]: ...
    def get_mutable_context(self) -> pydrake.systems.framework.Context_TAutoDiffXdU: ...
    def get_mutable_integrator(self) -> IntegratorBase_TAutoDiffXdU: ...
    def get_num_discrete_updates(self) -> int: ...
    def get_num_publishes(self) -> int: ...
    def get_num_steps_taken(self) -> int: ...
    def get_num_unrestricted_updates(self) -> int: ...
    def get_system(self) -> pydrake.systems.framework.System_TAutoDiffXdU: ...
    def get_target_realtime_rate(self) -> float: ...
    def has_context(self) -> bool: ...
    def reset_context(self, context: pydrake.systems.framework.Context_TAutoDiffXdU | None) -> None: ...
    def set_monitor(self, monitor: collections.abc.Callable[[pydrake.systems.framework.Context_TAutoDiffXdU], pydrake.systems.framework.EventStatus | None]) -> None: ...
    def set_target_realtime_rate(self, realtime_rate: float) -> None: ...

class _TemporaryName_N5drake7systems14IntegratorBaseINS_2ad8AutoDiffEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def GetStatisticsSummary(self) -> list[tuple[str, int | float]]: ...
    @overload
    def Initialize(self) -> None: ...
    @overload
    def Initialize(self) -> Any: ...
    @overload
    def Initialize(self) -> Any: ...
    def Reset(self) -> None: ...
    def ResetStatistics(self) -> None: ...
    def StartDenseIntegration(self) -> None: ...
    def StopDenseIntegration(self) -> pydrake.trajectories.PiecewisePolynomial_TAutoDiffXdU: ...
    def get_actual_initial_step_size_taken(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_context(self) -> pydrake.systems.framework.Context_TAutoDiffXdU: ...
    def get_dense_output(self) -> pydrake.trajectories.PiecewisePolynomial_TAutoDiffXdU: ...
    def get_fixed_step_mode(self) -> bool: ...
    def get_initial_step_size_target(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_largest_step_size_taken(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_maximum_step_size(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_mutable_context(self) -> pydrake.systems.framework.Context_TAutoDiffXdU: ...
    def get_num_derivative_evaluations(self) -> int: ...
    def get_num_step_shrinkages_from_error_control(self) -> int: ...
    def get_num_step_shrinkages_from_substep_failures(self) -> int: ...
    def get_num_steps_taken(self) -> int: ...
    def get_num_substep_failures(self) -> int: ...
    def get_requested_minimum_step_size(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_smallest_adapted_step_size_taken(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def get_target_accuracy(self) -> float: ...
    def get_throw_on_minimum_step_size_violation(self) -> bool: ...
    def request_initial_step_size_target(self, step_size: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def reset_context(self, context: pydrake.systems.framework.Context_TAutoDiffXdU | None = ...) -> None: ...
    def set_fixed_step_mode(self, flag: bool) -> None: ...
    @overload
    def set_maximum_step_size(self, max_step_size: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    @overload
    def set_maximum_step_size(self) -> Any: ...
    def set_requested_minimum_step_size(self, min_step_size: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_target_accuracy(self, accuracy: float) -> None: ...
    def set_throw_on_minimum_step_size_violation(self, throws: bool) -> None: ...

class _TemporaryName_N5drake7systems14IntegratorBaseINS_8symbolic10ExpressionEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def GetStatisticsSummary(self) -> list[tuple[str, int | float]]: ...
    @overload
    def Initialize(self) -> None: ...
    @overload
    def Initialize(self) -> Any: ...
    @overload
    def Initialize(self) -> Any: ...
    def Reset(self) -> None: ...
    def ResetStatistics(self) -> None: ...
    def StartDenseIntegration(self) -> None: ...
    def StopDenseIntegration(self) -> pydrake.trajectories.PiecewisePolynomial_TExpressionU: ...
    def get_actual_initial_step_size_taken(self) -> pydrake.symbolic.Expression: ...
    def get_context(self) -> pydrake.systems.framework.Context_TExpressionU: ...
    def get_dense_output(self) -> pydrake.trajectories.PiecewisePolynomial_TExpressionU: ...
    def get_fixed_step_mode(self) -> bool: ...
    def get_initial_step_size_target(self) -> pydrake.symbolic.Expression: ...
    def get_largest_step_size_taken(self) -> pydrake.symbolic.Expression: ...
    def get_maximum_step_size(self) -> pydrake.symbolic.Expression: ...
    def get_mutable_context(self) -> pydrake.systems.framework.Context_TExpressionU: ...
    def get_num_derivative_evaluations(self) -> int: ...
    def get_num_step_shrinkages_from_error_control(self) -> int: ...
    def get_num_step_shrinkages_from_substep_failures(self) -> int: ...
    def get_num_steps_taken(self) -> int: ...
    def get_num_substep_failures(self) -> int: ...
    def get_requested_minimum_step_size(self) -> pydrake.symbolic.Expression: ...
    def get_smallest_adapted_step_size_taken(self) -> pydrake.symbolic.Expression: ...
    def get_target_accuracy(self) -> float: ...
    def get_throw_on_minimum_step_size_violation(self) -> bool: ...
    def request_initial_step_size_target(self, step_size: pydrake.symbolic.Expression) -> None: ...
    def reset_context(self, context: pydrake.systems.framework.Context_TExpressionU | None = ...) -> None: ...
    def set_fixed_step_mode(self, flag: bool) -> None: ...
    @overload
    def set_maximum_step_size(self, max_step_size: pydrake.symbolic.Expression) -> None: ...
    @overload
    def set_maximum_step_size(self) -> Any: ...
    def set_requested_minimum_step_size(self, min_step_size: pydrake.symbolic.Expression) -> None: ...
    def set_target_accuracy(self, accuracy: float) -> None: ...
    def set_throw_on_minimum_step_size_violation(self, throws: bool) -> None: ...

class _TemporaryName_N5drake7systems21RungeKutta2IntegratorINS_2ad8AutoDiffEEE(IntegratorBase_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TAutoDiffXdU, max_step_size: pydrake.autodiffutils.AutoDiffXd, context: pydrake.systems.framework.Context_TAutoDiffXdU | None = ...) -> None: ...

class _TemporaryName_N5drake7systems21RungeKutta2IntegratorINS_8symbolic10ExpressionEEE(IntegratorBase_TExpressionU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TExpressionU, max_step_size: pydrake.symbolic.Expression, context: pydrake.systems.framework.Context_TExpressionU | None = ...) -> None: ...

class _TemporaryName_N5drake7systems21RungeKutta3IntegratorINS_2ad8AutoDiffEEE(IntegratorBase_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TAutoDiffXdU, context: pydrake.systems.framework.Context_TAutoDiffXdU | None = ...) -> None: ...

class _TemporaryName_N5drake7systems9SimulatorINS_2ad8AutoDiffEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TAutoDiffXdU, context: object | None = ...) -> None: ...
    def AdvancePendingEvents(self) -> SimulatorStatus: ...
    def AdvanceTo(self, boundary_time: pydrake.autodiffutils.AutoDiffXd, interruptible: bool = ...) -> SimulatorStatus: ...
    def Initialize(self, params: InitializeParams = ...) -> SimulatorStatus: ...
    def ResetStatistics(self) -> None: ...
    def clear_monitor(self) -> None: ...
    def get_actual_realtime_rate(self) -> float: ...
    def get_context(self) -> pydrake.systems.framework.Context_TAutoDiffXdU: ...
    def get_integrator(self) -> IntegratorBase_TAutoDiffXdU: ...
    def get_monitor(self) -> collections.abc.Callable[[pydrake.systems.framework.Context_TAutoDiffXdU], pydrake.systems.framework.EventStatus]: ...
    def get_mutable_context(self) -> pydrake.systems.framework.Context_TAutoDiffXdU: ...
    def get_mutable_integrator(self) -> IntegratorBase_TAutoDiffXdU: ...
    def get_num_discrete_updates(self) -> int: ...
    def get_num_publishes(self) -> int: ...
    def get_num_steps_taken(self) -> int: ...
    def get_num_unrestricted_updates(self) -> int: ...
    def get_system(self) -> pydrake.systems.framework.System_TAutoDiffXdU: ...
    def get_target_realtime_rate(self) -> float: ...
    def has_context(self) -> bool: ...
    def reset_context(self, context: pydrake.systems.framework.Context_TAutoDiffXdU | None) -> None: ...
    def set_monitor(self, monitor: collections.abc.Callable[[pydrake.systems.framework.Context_TAutoDiffXdU], pydrake.systems.framework.EventStatus | None]) -> None: ...
    def set_target_realtime_rate(self, realtime_rate: float) -> None: ...



"""
MeshCat LCM Display Server (MeLDiS)

A standalone program that can display Drake visualizations in MeshCat
by listening for LCM messages that are broadcast by the simulation.

From a Drake source build, run this as::

  bazel run //tools:meldis &

From a Drake binary release (including pip releases), run this as::

  python3 -m pydrake.visualization.meldis

For binary releases (except for pip) there is also a shortcut available as::

  /opt/drake/bin/meldis

In many cases, passing ``-w`` (i.e., ``--open-window``) to the program will be
convenient::

  bazel run //tools:meldis -- -w &
"""

import argparse
import os
import sys
import webbrowser

from pydrake.common import configure_logging as _configure_logging
from pydrake.common.yaml import yaml_load_typed as _yaml_load_typed
from pydrake.visualization._meldis import _DEFAULT_MESHCAT_PARAMS
from pydrake.visualization._meldis import Meldis as _Meldis


def _available_browsers():
    """Returns the list of known webbrowser controller names. There does not
    appear to be a public API for this, so we best-effort call a private one.
    """
    try:
        webbrowser.register_standard_browsers()
        return sorted(webbrowser._browsers)
    except Exception:
        return []


def _main(args=None):
    # Make cwd be what the user expected, not the runfiles tree.
    if "BUILD_WORKING_DIRECTORY" in os.environ:
        os.chdir(os.environ["BUILD_WORKING_DIRECTORY"])
    _configure_logging()
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--host",
        action="store",
        help="The http listen host for MeshCat. If none is given, 'localhost'"
        " will be used by default. In any case, the result will be printed to"
        " the console.",
    )
    parser.add_argument(
        "-p",
        "--port",
        action="store",
        metavar="NUM",
        type=int,
        help="The http listen port for MeshCat. If none is given, a default"
        " will be chosen and printed to the console.",
    )
    parser.add_argument(
        "-t",
        "--open-tab",
        dest="browser_new",
        action="store_const",
        const=2,
        default=None,
        help="Open the MeshCat display in a browser tab.",
    )
    parser.add_argument(
        "-w",
        "--open-window",
        dest="browser_new",
        action="store_const",
        const=1,
        default=None,
        help="Open the MeshCat display in a new browser window.",
    )
    parser.add_argument(
        "--browser",
        metavar="NAME",
        choices=_available_browsers(),
        help="Open the MeshCat display using the given browser. "
        "By default, opens as a new window (use --open-tab to override). "
        "When no --browser is provided, the --open-tab or --open-window flags "
        "use the $BROWSER environment variable by default. "
        "(Available names: %(choices)s)",
    )
    parser.add_argument(
        "--idle-timeout",
        metavar="TIME",
        type=float,
        default=15 * 60,
        help="When no web browser has been connected for this many seconds,"
        " this program will automatically exit. Set to 0 to run indefinitely.",
    )
    parser.add_argument(
        "--meshcat-params",
        metavar="PATH",
        help="Filesystem path to a YAML or JSON config for MeshcatParams. "
        "This can be used to configure Meshcat's initial properties. "
        "For options that are available as both command line arguments and "
        "YAML params (e.g., --port), the command line takes precedence.",
    )
    parser.add_argument(
        "--environment_map",
        metavar="PATH",
        help="Filesystem path to an image to be used as an environment map. "
        "It must be an image type normally used by your browser (e.g., "
        ".jpg, .png, etc.). HDR images are not supported yet.",
    )
    args = parser.parse_args(args)
    meshcat_params = None
    if args.meshcat_params is not None:
        meshcat_params = _yaml_load_typed(
            filename=args.meshcat_params, defaults=_DEFAULT_MESHCAT_PARAMS
        )
    meldis = _Meldis(
        meshcat_host=args.host,
        meshcat_port=args.port,
        meshcat_params=meshcat_params,
        environment_map=args.environment_map,
    )
    if args.browser is not None and args.browser_new is None:
        args.browser_new = 1
    if args.browser_new is not None:
        url = meldis.meshcat.web_url()
        controller = webbrowser.get(args.browser)
        controller.open(url=url, new=args.browser_new)
    idle_timeout = args.idle_timeout
    if idle_timeout == 0.0:
        idle_timeout = None
    elif idle_timeout < 0.0:
        parser.error("The --idle_timeout cannot be negative.")
    try:
        meldis.serve_forever(idle_timeout=idle_timeout)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    _main(args=sys.argv[1:])



import pydrake.multibody.plant
import pydrake.systems.framework

class MultibodyPositionToGeometryPose(pydrake.systems.framework.LeafSystem):
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant, input_multibody_state: bool = ...) -> None: ...



import enum
import pydrake.autodiffutils
from pydrake.common.cpp_template import DeformableBodyConfig_ as DeformableBodyConfig_, ForceDensityFieldBase_ as ForceDensityFieldBase_
from typing import Callable, ClassVar

class DeformableBodyConfig:
    def __init__(self) -> None: ...
    def mass_damping_coefficient(self) -> float: ...
    def mass_density(self) -> float: ...
    def material_model(self) -> MaterialModel: ...
    def poissons_ratio(self) -> float: ...
    def set_mass_damping_coefficient(self, mass_damping_coefficient: float) -> None: ...
    def set_mass_density(self, mass_density: float) -> None: ...
    def set_material_model(self, material_model: MaterialModel) -> None: ...
    def set_poissons_ratio(self, poissons_ratio: float) -> None: ...
    def set_stiffness_damping_coefficient(self, stiffness_damping_coefficient: float) -> None: ...
    def set_youngs_modulus(self, youngs_modulus: float) -> None: ...
    def stiffness_damping_coefficient(self) -> float: ...
    def youngs_modulus(self) -> float: ...
    def __copy__(self) -> DeformableBodyConfig: ...
    def __deepcopy__(self, arg: dict) -> DeformableBodyConfig: ...

class DeformableBodyConfig_𝓣AutoDiffXd𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...
    def mass_damping_coefficient(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def mass_density(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def material_model(self) -> MaterialModel: ...
    def poissons_ratio(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def set_mass_damping_coefficient(self, mass_damping_coefficient: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_mass_density(self, mass_density: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_material_model(self, material_model: MaterialModel) -> None: ...
    def set_poissons_ratio(self, poissons_ratio: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_stiffness_damping_coefficient(self, stiffness_damping_coefficient: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_youngs_modulus(self, youngs_modulus: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def stiffness_damping_coefficient(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def youngs_modulus(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def __copy__(self) -> DeformableBodyConfig_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> DeformableBodyConfig_TAutoDiffXdU: ...

class ForceDensityFieldBase:
    def __init__(self, *args, **kwargs) -> None: ...
    def Clone(self) -> ForceDensityFieldBase: ...
    def EvaluateAt(self, *args, **kwargs): ...
    def density_type(self) -> ForceDensityType: ...
    def __copy__(self) -> ForceDensityFieldBase: ...
    def __deepcopy__(self, arg: dict) -> ForceDensityFieldBase: ...

class ForceDensityFieldBase_𝓣AutoDiffXd𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def Clone(self) -> ForceDensityFieldBase_TAutoDiffXdU: ...
    def EvaluateAt(self, *args, **kwargs): ...
    def density_type(self) -> ForceDensityType: ...
    def __copy__(self) -> ForceDensityFieldBase_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> ForceDensityFieldBase_TAutoDiffXdU: ...

class ForceDensityFieldBase_𝓣Expression𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def Clone(self) -> ForceDensityFieldBase_TExpressionU: ...
    def EvaluateAt(self, *args, **kwargs): ...
    def density_type(self) -> ForceDensityType: ...
    def __copy__(self) -> ForceDensityFieldBase_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> ForceDensityFieldBase_TExpressionU: ...

class ForceDensityType(enum.Enum):
    __new__: ClassVar[Callable] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _hashable_values_: ClassVar[list] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[object]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _unhashable_values_map_: ClassVar[dict] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    _value_repr_: ClassVar[None] = ...
    kPerCurrentVolume: ClassVar[ForceDensityType] = ...
    kPerReferenceVolume: ClassVar[ForceDensityType] = ...
    __nb_enum__: ClassVar[PyCapsule] = ...

class MaterialModel(enum.Enum):
    __new__: ClassVar[Callable] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _hashable_values_: ClassVar[list] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[object]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _unhashable_values_map_: ClassVar[dict] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    _value_repr_: ClassVar[None] = ...
    kCorotated: ClassVar[MaterialModel] = ...
    kLinear: ClassVar[MaterialModel] = ...
    kLinearCorotated: ClassVar[MaterialModel] = ...
    kNeoHookean: ClassVar[MaterialModel] = ...
    __nb_enum__: ClassVar[PyCapsule] = ...

class _TemporaryName_N5drake9multibody21ForceDensityFieldBaseINS_2ad8AutoDiffEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def Clone(self) -> ForceDensityFieldBase_TAutoDiffXdU: ...
    def EvaluateAt(self, *args, **kwargs): ...
    def density_type(self) -> ForceDensityType: ...
    def __copy__(self) -> ForceDensityFieldBase_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> ForceDensityFieldBase_TAutoDiffXdU: ...

class _TemporaryName_N5drake9multibody21ForceDensityFieldBaseINS_8symbolic10ExpressionEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def Clone(self) -> ForceDensityFieldBase_TExpressionU: ...
    def EvaluateAt(self, *args, **kwargs): ...
    def density_type(self) -> ForceDensityType: ...
    def __copy__(self) -> ForceDensityFieldBase_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> ForceDensityFieldBase_TExpressionU: ...

class _TemporaryName_N5drake9multibody3fem20DeformableBodyConfigINS_2ad8AutoDiffEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...
    def mass_damping_coefficient(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def mass_density(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def material_model(self) -> MaterialModel: ...
    def poissons_ratio(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def set_mass_damping_coefficient(self, mass_damping_coefficient: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_mass_density(self, mass_density: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_material_model(self, material_model: MaterialModel) -> None: ...
    def set_poissons_ratio(self, poissons_ratio: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_stiffness_damping_coefficient(self, stiffness_damping_coefficient: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_youngs_modulus(self, youngs_modulus: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def stiffness_damping_coefficient(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def youngs_modulus(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def __copy__(self) -> DeformableBodyConfig_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> DeformableBodyConfig_TAutoDiffXdU: ...



import collections.abc
import nanobind
import pydrake.common.value
import pydrake.lcm
import pydrake.systems.framework
from typing import Callable, ClassVar, overload

ApplyLcmBusConfig: nanobind.nb_func

class LcmBuses:
    def __init__(self) -> None: ...
    def Add(self, bus_name: str, bus: pydrake.lcm.DrakeLcmInterface) -> None: ...
    def Find(self, description_of_caller: str, bus_name: str) -> pydrake.lcm.DrakeLcmInterface: ...
    def GetAllBusNames(self) -> list[str]: ...
    def size(self) -> int: ...
    @property
    def kLcmUrlMemqNull(self): ...

class LcmInterfaceSystem(pydrake.systems.framework.LeafSystem):
    def __init__(self, lcm: pydrake.lcm.DrakeLcmInterface) -> None: ...
    def HandleSubscriptions(self, timeout_millis: int) -> int: ...
    def get_lcm_url(self) -> str: ...

class LcmPublisherSystem(pydrake.systems.framework.LeafSystem):
    Make: ClassVar[Callable] = ...
    @overload
    def __init__(self, channel: str, serializer: SerializerInterface, lcm: LcmInterfaceSystem | None, publish_period: float = ..., publish_offset: float = ...) -> None: ...
    @overload
    def __init__(self, channel: str, serializer: SerializerInterface, lcm: pydrake.lcm.DrakeLcmInterface, publish_period: float = ..., publish_offset: float = ...) -> None: ...
    @overload
    def __init__(self, channel: str, serializer: SerializerInterface, lcm: LcmInterfaceSystem, publish_triggers: collections.abc.Set[pydrake.systems.framework.TriggerType], publish_period: float = ..., publish_offset: float = ...) -> None: ...
    @overload
    def __init__(self, channel: str, serializer: SerializerInterface, lcm: pydrake.lcm.DrakeLcmInterface, publish_triggers: collections.abc.Set[pydrake.systems.framework.TriggerType], publish_period: float = ..., publish_offset: float = ...) -> None: ...

class LcmScopeSystem(pydrake.systems.framework.LeafSystem):
    AddToBuilder: ClassVar[nanobind.nb_func] = ...
    def __init__(self, size: int) -> None: ...

class LcmSubscriberSystem(pydrake.systems.framework.LeafSystem):
    Make: ClassVar[Callable] = ...
    @overload
    def __init__(self, channel: str, serializer: SerializerInterface, lcm: LcmInterfaceSystem, wait_for_message_on_initialization_timeout: float = ...) -> None: ...
    @overload
    def __init__(self, channel: str, serializer: SerializerInterface, lcm: pydrake.lcm.DrakeLcmInterface, wait_for_message_on_initialization_timeout: float = ...) -> None: ...
    def WaitForMessage(self, old_message_count: int, message: pydrake.common.value.AbstractValue | None = ..., timeout: float = ...) -> int: ...

class PySerializer(SerializerInterface):
    __init__: ClassVar[Callable] = ...
    CreateDefaultValue: ClassVar[Callable] = ...
    Deserialize: ClassVar[Callable] = ...
    Serialize: ClassVar[Callable] = ...

class SerializerInterface:
    def __init__(self) -> None: ...
    def CreateDefaultValue(self) -> pydrake.common.value.AbstractValue: ...
    def Deserialize(self, message_bytes: bytes, abstract_value: pydrake.common.value.AbstractValue) -> None: ...
    def Serialize(self, abstract_value: pydrake.common.value.AbstractValue) -> bytes: ...

class _Serializer(SerializerInterface):
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_acrobot_x𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_acrobot_y𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_allegro_command𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_allegro_status𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_call_python_data𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_call_python𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_contact_results_for_viz𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_drake_signal𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_force_torque𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_header𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_hydroelastic_contact_surface_for_viz𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_hydroelastic_quadrature_per_point_data_for_viz𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_iiwa_command𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_iiwa_status_telemetry𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_iiwa_status𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_image_array𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_image𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_jaco_command𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_jaco_status𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_panda_command𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_panda_status𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_planar_gripper_command𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_planar_gripper_finger_command𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_planar_gripper_finger_face_assignments𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_planar_gripper_finger_face_assignment𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_planar_gripper_finger_status𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_planar_gripper_status𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_planar_manipuland_status𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_planar_plant_state𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_point_cloud_field𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_point_cloud𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_point_pair_contact_info_for_viz𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_point𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_quaternion𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_robot_plan𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_robot_state𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_schunk_wsg_command𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_schunk_wsg_status𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_scope𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_viewer_command𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_viewer_draw𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_viewer_geometry_data𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_viewer_link_data𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _Serializer_𝓣lcmt_viewer_load_robot𝓤(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_10lcmt_imageEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_10lcmt_pointEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_10lcmt_scopeEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_11lcmt_headerEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_14lcmt_acrobot_xEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_14lcmt_acrobot_yEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_15lcmt_quaternionEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_15lcmt_robot_planEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_16lcmt_call_pythonEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_16lcmt_iiwa_statusEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_16lcmt_image_arrayEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_16lcmt_jaco_statusEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_16lcmt_point_cloudEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_16lcmt_robot_stateEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_16lcmt_viewer_drawEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_17lcmt_drake_signalEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_17lcmt_force_torqueEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_17lcmt_iiwa_commandEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_17lcmt_jaco_commandEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_17lcmt_panda_statusEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_18lcmt_panda_commandEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_19lcmt_allegro_statusEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_19lcmt_viewer_commandEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_20lcmt_allegro_commandEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_21lcmt_call_python_dataEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_21lcmt_viewer_link_dataEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_22lcmt_point_cloud_fieldEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_22lcmt_schunk_wsg_statusEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_22lcmt_viewer_load_robotEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_23lcmt_planar_plant_stateEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_23lcmt_schunk_wsg_commandEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_25lcmt_viewer_geometry_dataEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_26lcmt_iiwa_status_telemetryEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_26lcmt_planar_gripper_statusEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_27lcmt_planar_gripper_commandEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_28lcmt_contact_results_for_vizEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_29lcmt_planar_manipuland_statusEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_33lcmt_planar_gripper_finger_statusEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_34lcmt_planar_gripper_finger_commandEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_36lcmt_point_pair_contact_info_for_vizEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_41lcmt_hydroelastic_contact_surface_for_vizEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_42lcmt_planar_gripper_finger_face_assignmentEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_43lcmt_planar_gripper_finger_face_assignmentsEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...

class _TemporaryName_N5drake7systems3lcm10SerializerINS_51lcmt_hydroelastic_quadrature_per_point_data_for_vizEEE(SerializerInterface):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self) -> None: ...



"""Drake offers Python tutorials that can be previewed and executed as Jupyter
notebooks online with no need for local installation. To run Drake's tutorials
online, refer to the `Drake Tutorials <https://drake.mit.edu/>`_ website.

Alternatively, to run Drake's tutorials locally from an
`installed <https://drake.mit.edu/installation.html>`_ copy of Drake,
run ``python3 -m pydrake.tutorials`` to launch a Jupyter browser.

Be sure your ``PYTHONPATH`` has been set per the installation instructions,
e.g., via ``source env/bin/activate`` in the
`pip instructions <https://drake.mit.edu/pip.html>`_.

If you haven't done so already, you'll also need to install the Jupyter
notebook package on your system:

- For pip or macOS, use: ``pip install notebook``.
- For Ubuntu, use ``sudo apt-get install jupyter-notebook``.
"""


def __main():
    import sys

    import pydrake

    sys.argv = ["notebook", f"{pydrake.getDrakePath()}/tutorials/index.ipynb"]
    imported = False
    try:
        # Try the Jupyter >= 7 spelling first.
        from notebook import app

        imported = True
    except ImportError:
        pass
    if not imported:
        try:
            # Try the Jupyter < 7 spelling as a fallback.
            from notebook import notebookapp as app

            imported = True
        except ImportError:
            pass
    if not imported:
        print("ERROR: the Jupyter notebook runtime is not installed!")
        print()
        print(__doc__)
        sys.exit(1)
    app.launch_new_instance()


if __name__ == "__main__":
    __main()



# ruff: noqa: F401 (unused-import)

from ._meldis import (
    Meldis,
)
from ._model_visualizer import (
    AddFrameTriadIllustration,
    ModelVisualizer,
)
from ._plotting import (
    plot_sublevelset_expression,
    plot_sublevelset_quadratic,
)
from ._video import (
    VideoWriter,
)

__all__ = [x for x in globals() if not x.startswith("_")]



# ruff: noqa: F821 (undefined-name). This file is only a fragment.

import pydrake.autodiffutils as _ad
from pydrake.common import (
    _MangledName,
)
from pydrake.common import (
    pretty_class_name as _pretty_class_name,
)
import pydrake.symbolic as _sym


def _indented_repr(o):
    """Returns repr(o), with any lines beyond the first one indented +2."""
    return repr(o).replace("\n", "\n  ")


def _spatial_vector_repr(rotation_name, translation_name):
    def repr_with_closure(self):
        rotation = self.rotational().tolist()
        translation = self.translational().tolist()
        return (
            f"{_pretty_class_name(type(self))}(\n"
            f"  {rotation_name}={_indented_repr(rotation)},\n"
            f"  {translation_name}={_indented_repr(translation)},\n"
            f")"
        )

    return repr_with_closure


def _add_repr_functions():
    for T in [float, _ad.AutoDiffXd, _sym.Expression]:
        SpatialVelocity_[T].__repr__ = _spatial_vector_repr("w", "v")
        SpatialMomentum_[T].__repr__ = _spatial_vector_repr("h", "l")
        SpatialAcceleration_[T].__repr__ = _spatial_vector_repr("alpha", "a")
        SpatialForce_[T].__repr__ = _spatial_vector_repr("tau", "f")


_add_repr_functions()


def __getattr__(name):
    """Rewrites requests for Foo[bar] into their mangled form, for backwards
    compatibility with unpickling.
    """
    return _MangledName.module_getattr(
        module_name=__name__, module_globals=globals(), name=name
    )



import enum
import nanobind
import pydrake.multibody.plant
import pydrake.systems.analysis
import pydrake.systems.framework
import pydrake.trajectories
from _typeshed import Incomplete
from typing import Any, Callable, ClassVar, overload

DiscreteTimeLinearQuadraticRegulator: nanobind.nb_func
FiniteHorizonLinearQuadraticRegulator: nanobind.nb_func
FittedValueIteration: nanobind.nb_func
LinearProgrammingApproximateDynamicProgramming: nanobind.nb_func
LinearQuadraticRegulator: nanobind.nb_func
MakeFiniteHorizonLinearQuadraticRegulator: nanobind.nb_func

class DynamicProgrammingOptions:
    class PeriodicBoundaryCondition:
        high: Incomplete
        low: Incomplete
        state_index: Incomplete
        def __init__(self, state_index: int, low: float, high: float) -> None: ...
    assume_non_continuous_states_are_fixed: Incomplete
    convergence_tol: float
    discount_factor: float
    input_port_index: pydrake.systems.framework.InputPortSelection | pydrake.systems.framework.InputPortIndex
    periodic_boundary_conditions: Incomplete
    visualization_callback: Incomplete
    def __init__(self) -> None: ...

class FiniteHorizonLinearQuadraticRegulatorOptions:
    N: Incomplete
    Qf: Incomplete
    input_port_index: pydrake.systems.framework.InputPortSelection | pydrake.systems.framework.InputPortIndex
    simulator_config: pydrake.systems.analysis.SimulatorConfig
    u0: pydrake.trajectories.Trajectory
    ud: Incomplete
    use_square_root_method: bool
    x0: pydrake.trajectories.Trajectory
    xd: Any
    def __init__(self) -> None: ...

class FiniteHorizonLinearQuadraticRegulatorResult:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def K(self) -> pydrake.trajectories.Trajectory: ...
    @property
    def S(self): ...
    @property
    def k0(self): ...
    @property
    def s0(self): ...
    @property
    def sx(self): ...
    @property
    def u0(self): ...
    @property
    def x0(self): ...

class InverseDynamics(pydrake.systems.framework.LeafSystem):
    class InverseDynamicsMode(enum.Enum):
        __new__: ClassVar[Callable] = ...
        _generate_next_value_: ClassVar[Callable] = ...
        _hashable_values_: ClassVar[list] = ...
        _member_map_: ClassVar[dict] = ...
        _member_names_: ClassVar[list] = ...
        _member_type_: ClassVar[type[object]] = ...
        _unhashable_values_: ClassVar[list] = ...
        _unhashable_values_map_: ClassVar[dict] = ...
        _use_args_: ClassVar[bool] = ...
        _value2member_map_: ClassVar[dict] = ...
        _value_repr_: ClassVar[None] = ...
        kGravityCompensation: ClassVar[InverseDynamics.InverseDynamicsMode] = ...
        kInverseDynamics: ClassVar[InverseDynamics.InverseDynamicsMode] = ...
        __nb_enum__: ClassVar[PyCapsule] = ...
    kGravityCompensation: ClassVar[InverseDynamics.InverseDynamicsMode] = ...
    kInverseDynamics: ClassVar[InverseDynamics.InverseDynamicsMode] = ...
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant, mode: InverseDynamics.InverseDynamicsMode = ..., plant_context: pydrake.systems.framework.Context | None = ...) -> None: ...
    def get_input_port_desired_acceleration(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_estimated_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_output_port_generalized_force(self) -> pydrake.systems.framework.OutputPort: ...
    def is_pure_gravity_compensation(self) -> bool: ...

class InverseDynamicsController(pydrake.systems.framework.Diagram):
    def __init__(self, robot: pydrake.multibody.plant.MultibodyPlant, kp, ki, kd, has_reference_acceleration: bool, plant_context: pydrake.systems.framework.Context | None = ...) -> None: ...
    def get_input_port_desired_acceleration(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_desired_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_estimated_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_multibody_plant_for_control(self) -> pydrake.multibody.plant.MultibodyPlant: ...
    def get_output_port_control(self) -> pydrake.systems.framework.OutputPort: ...
    def get_output_port_generalized_force(self) -> pydrake.systems.framework.OutputPort: ...
    def set_integral_value(self, arg0: pydrake.systems.framework.Context, arg1) -> None: ...

class JointStiffnessController(pydrake.systems.framework.LeafSystem):
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant, kp, kd) -> None: ...
    def get_input_port_desired_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_estimated_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_multibody_plant(self) -> pydrake.multibody.plant.MultibodyPlant: ...
    def get_output_port_actuation(self) -> pydrake.systems.framework.OutputPort: ...

class PeriodicBoundaryCondition:
    high: Incomplete
    low: Incomplete
    state_index: Incomplete
    def __init__(self, state_index: int, low: float, high: float) -> None: ...

class PidControlledSystem(pydrake.systems.framework.Diagram):
    @overload
    def __init__(self, plant: pydrake.systems.framework.System, kp: float, ki: float, kd: float, state_output_port_index: int = ..., plant_input_port_index: int = ...) -> None: ...
    @overload
    def __init__(self, plant: pydrake.systems.framework.System, kp, ki, kd, state_output_port_index: int = ..., plant_input_port_index: int = ...) -> None: ...
    @overload
    def __init__(self, plant: pydrake.systems.framework.System, feedback_selector, kp: float, ki: float, kd: float, state_output_port_index: int = ..., plant_input_port_index: int = ...) -> None: ...
    @overload
    def __init__(self, plant: pydrake.systems.framework.System, feedback_selector, kp, ki, kd, state_output_port_index: int = ..., plant_input_port_index: int = ...) -> None: ...
    def get_control_input_port(self) -> pydrake.systems.framework.InputPort: ...
    def get_state_input_port(self) -> pydrake.systems.framework.InputPort: ...
    def get_state_output_port(self) -> pydrake.systems.framework.OutputPort: ...

class PidController(pydrake.systems.framework.LeafSystem):
    @overload
    def __init__(self, kp, ki, kd) -> None: ...
    @overload
    def __init__(self, state_projection, kp, ki, kd) -> None: ...
    @overload
    def __init__(self, state_projection, output_projection, kp, ki, kd) -> None: ...
    def get_Kd_vector(self, *args, **kwargs): ...
    def get_Ki_vector(self, *args, **kwargs): ...
    def get_Kp_vector(self, *args, **kwargs): ...
    def get_input_port_desired_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_estimated_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_output_port_control(self) -> pydrake.systems.framework.OutputPort: ...



# ruff: noqa: F401,F403 (unused-import, import-star)
# ruff: isort: skip_file

from . import *
from .compatibility import *
from .containers import *

# - `cpp_param` does not offer public Drake symbols.
# - `cpp_template` does not offer public Drake symbols.
# - `deprecation` does not offer public Drake symbols.

from .eigen_geometry import *
from .schema import *
from .yaml import *
from .value import *



import copy

import numpy as np

from pydrake.common.value import Value
from pydrake.geometry import (
    ClippingRange,
    DepthRange,
    DepthRenderCamera,
    MakeRenderEngineVtk,
    RenderCameraCore,
    RenderEngineVtkParams,
    SceneGraph,
)
from pydrake.math import RigidTransform
from pydrake.systems.framework import LeafSystem
from pydrake.systems.sensors import (
    CameraInfo,
    ImageRgba8U,
    RgbdSensor,
)
from pydrake.visualization import (
    ColorizeDepthImage,
    ColorizeLabelImage,
    ConcatenateImages,
)


class VideoWriter(LeafSystem):
    """Publishes RgbdSensor output to a video file.

    .. pydrake_system::

        name: VideoWriter
        input_ports:
        - color_image

    The video will contain recorded images from the color_image input port.

    The methods AddToBuilder() or ConnectToRgbdSensor() make it easy to record
    color and/or depth and/or labels all at once.  For companion systems, see
    also ColorizeDepthImage, ColorizeLabelImage, and ConcatenateImages.

    This class delegates actual video output to one of two video backends,
    either "PIL" (aka Pillow) or "cv2". PIL generally only supports image
    formats (e.g., gif, apng, webp) while cv2 also supports movies (e.g., mp4).

    Warning:
        Once all images have been published, you must call
        ``video_writer.Save()`` to finish writing to the video file.

    Warning:
        This class will fail at construction time if the specified ``backend``
        module cannot be imported. You must ensure that whichever backend you
        choose is available in your environment. Drake neither bundles nor
        depends on either one.
    """

    def __init__(self, *, filename, fps=16.0, backend="PIL", fourcc=None):
        """Constructs a VideoWriter system.

        In many cases, the AddToBuilder() or ConnectRgbdSensor() methods might
        be easier to use than this constructor.

        Args:
            filename: filename to write, e.g., ``"output.gif"`` or
                ``"output.mp4"``
            fps: the output video's frame rate (in frames per second)
            backend: which backend to use: "PIL" or "cv2"
            fourcc: when using the cv2 backend, which encoder to use;
                good choices are "mp4v" or "avc1"; defaults to "mp4v";
                refer to the OpenCV documentation for details.
        """
        LeafSystem.__init__(self)
        self._filename = filename
        self._fps = fps
        self._input = self.DeclareAbstractInputPort(
            name="color_image", model_value=Value(ImageRgba8U())
        )
        # TODO(jwnimmer-tri) Support forced triggers as well (so users can
        # manually record videos of prescribed motion).
        self.DeclarePeriodicPublishEvent(1.0 / fps, 0.0, self._publish)
        self._cv2_writer = None
        self._pil_images = None
        if backend == "PIL":
            from PIL import Image

            self._backend = Image
            self._write = self._write_pil
        elif backend == "cv2":
            import cv2

            self._backend = cv2
            self._write = self._write_cv2
            self._fourcc = fourcc or "mp4v"
            if len(self._fourcc) != 4:
                raise ValueError(f"The fourcc={fourcc!r} must be 4 characters")
        else:
            raise RuntimeError(f"Invalid backend={backend!r}")

    @staticmethod
    def AddToBuilder(
        *,
        filename,
        builder,
        sensor_pose,
        fps=16.0,
        width=320,
        height=240,
        fov_y=np.pi / 6,
        near=0.01,
        far=10.0,
        kinds=None,
        backend="PIL",
        fourcc=None,
    ):
        """Adds a RgbdSensor and VideoWriter system to the given builder, using
        a world-fixed pose. Returns the VideoWriter system.

        See also ConnectRgbdSensor() in case you want to attach a VideoWriter
        to an already existing RgbdSensor (e.g., on attached to a robot).

        Args:
            filename: filename to write, e.g., ``"output.gif"`` or
                ``"output.mp4"``
            builder: the DiagramBuilder
            sensor_pose: the world-fixed position for the video camera
            fps: the output video's frame rate (in frames per second)
            width: video camera width (in pixels)
            height: video camera width (in pixels)
            fov_y: video camera fov (in radians)
            near: clipping plane distance (in meters)
            far: clipping plane distance (in meters)
            kinds: which image kind(s) to include in the video; valid
                options are ``"color"``, ``"label"``, and/or ``"depth"``
            backend: which backend to use: "PIL" or "cv2".
            fourcc: when using the cv2 backend, which encoder to use;
                good choices are "mp4v" or "avc1"; defaults to "mp4v"
                refer to the OpenCV documentation for details.

        Warning:
            Once all images have been published, you must call
            ``video_writer.Save()`` to finish writing to the video file.
        """
        sensor = VideoWriter._AddRgbdSensor(
            builder=builder,
            pose=sensor_pose,
            width=width,
            height=height,
            fov_y=fov_y,
            near=near,
            far=far,
        )
        writer = VideoWriter(
            filename=filename, fps=fps, backend=backend, fourcc=fourcc
        )
        builder.AddSystem(writer)
        writer.ConnectRgbdSensor(builder=builder, sensor=sensor, kinds=kinds)
        return writer

    @staticmethod
    def _AddRgbdSensor(*, builder, pose, width, height, fov_y, near, far):
        """Helper function that adds a fixed-pose RgbdSensor to a scene.
        Returns the sensor system, already added to the builder and connected
        to the scene graph and configured to use the VTK render engine.
        """
        scene_graph = [
            x for x in builder.GetSystems() if x.get_name() == "scene_graph"
        ][0]
        if not scene_graph.HasRenderer("vtk"):
            scene_graph.AddRenderer(
                "vtk", MakeRenderEngineVtk(RenderEngineVtkParams())
            )
        intrinsics = CameraInfo(width, height, fov_y)
        clip = ClippingRange(near, far)
        camera = DepthRenderCamera(
            RenderCameraCore("vtk", intrinsics, clip, RigidTransform()),
            DepthRange(near, far),
        )
        sensor = RgbdSensor(SceneGraph.world_frame_id(), pose, camera)
        builder.AddSystem(sensor)
        builder.Connect(
            scene_graph.GetOutputPort("query"),
            sensor.GetInputPort("geometry_query"),
        )
        return sensor

    def ConnectRgbdSensor(self, *, builder, sensor, kinds=None):
        """Adds a VideoWriter system to the given builder and connects it to
        the given RgbdSensor. Returns the VideoWriter system.

        See also AddToBuilder() in case you want to record video from a world-
        fixed pose by creating a new sensor.

        Args:
            builder: the DiagramBuilder
            kinds: which image kind(s) to include in the video; valid
                options are ``"color"``, ``"label"``, and/or ``"depth"``;
                when set to ``None``, defaults to ``"color"`` only.

        Warning:
            Once all images have been published, you must call
            ``video_writer.Save()`` to finish writing to the video file.
        """
        # Make a list of ImageRgba8U output ports to feed as video input.
        image_sources = []
        for kind in kinds or ("color",):
            if kind == "color":
                image_sources.append(sensor.GetOutputPort("color_image"))
            elif kind == "depth":
                converter = builder.AddSystem(ColorizeDepthImage())
                builder.Connect(
                    sensor.GetOutputPort("depth_image_32f"),
                    converter.GetInputPort("depth_image_32f"),
                )
                image_sources.append(converter.get_output_port())
            elif kind == "label":
                converter = builder.AddSystem(ColorizeLabelImage())
                builder.Connect(
                    sensor.GetOutputPort("label_image"),
                    converter.get_input_port(),
                )
                image_sources.append(converter.get_output_port())
            else:
                raise RuntimeError(f"Unknown image kind={kind!r}")
        num_sources = len(image_sources)
        if num_sources == 1:
            image_source = image_sources[0]
        else:
            stacker = builder.AddSystem(ConcatenateImages(cols=num_sources))
            for i, source in enumerate(image_sources):
                builder.Connect(source, stacker.get_input_port(row=0, col=i))
            image_source = stacker.get_output_port()
        builder.Connect(image_source, self.get_input_port())

    def Save(self):
        """Flushes all images to the video file and closes the file.

        Warning:
            Continuing a simulation after calling Save() will begin to
            overwrite the prior video with a new one.
        """
        # For PIL.
        if self._pil_images is not None:
            images = self._pil_images
            frame_millis = int(1000.0 / self._fps)
            images[0].save(
                self._filename,
                save_all=True,
                append_images=images[1:],
                optimize=True,
                duration=frame_millis,
            )
        self._pil_images = None
        # For cv2.
        if self._cv2_writer is not None:
            self._cv2_writer.release()
        self._cv2_writer = None

    def _publish(self, context):
        """The framework event handler that saves one input image."""
        color = self._input.Eval(context)
        # Call the backend-specific function that was set by our constructor.
        self._write(rgba=color.data)

    def _write_pil(self, *, rgba):
        """Saves one input image (when we're configured to use PIL)."""
        # Grab the `from PIL import Image` that we stored at construction-time.
        Image = self._backend
        image = Image.fromarray(copy.copy(rgba), mode="RGBA")
        if self._pil_images is None:
            self._pil_images = [image]
        else:
            self._pil_images.append(image)

    def _write_cv2(self, *, rgba):
        """Saves one input image (when we're configured to use cv2)."""
        # Grab `import cv2` that we stored at construction-time.
        cv2 = self._backend
        # Open the output file upon the first publish event.
        if self._cv2_writer is None:
            fourcc = cv2.VideoWriter.fourcc(*self._fourcc)
            (height, width, _) = rgba.shape
            self._cv2_writer = cv2.VideoWriter(
                self._filename, fourcc, self._fps, (width, height)
            )
        bgra = cv2.cvtColor(rgba, cv2.COLOR_RGB2BGR)
        self._cv2_writer.write(bgra)



import nanobind
import pydrake.systems.framework
from pydrake.common.cpp_template import LuenbergerObserver_ as LuenbergerObserver_
from typing import ClassVar

DiscreteTimeSteadyStateKalmanFilter: nanobind.nb_func
SteadyStateKalmanFilter: nanobind.nb_func

class LuenbergerObserver(pydrake.systems.framework.LeafSystem):
    def __init__(self, observed_system: pydrake.systems.framework.System, observed_system_context: pydrake.systems.framework.Context, observer_gain) -> None: ...
    def L(self, *args, **kwargs): ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort: ...
    def observer_gain(self, *args, **kwargs): ...

class LuenbergerObserver_𝓣AutoDiffXd𝓤(pydrake.systems.framework.LeafSystem_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, observed_system: pydrake.systems.framework.System_TAutoDiffXdU, observed_system_context: pydrake.systems.framework.Context_TAutoDiffXdU, observer_gain) -> None: ...
    def L(self, *args, **kwargs): ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort_TAutoDiffXdU: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort_TAutoDiffXdU: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort_TAutoDiffXdU: ...
    def observer_gain(self, *args, **kwargs): ...

class LuenbergerObserver_𝓣Expression𝓤(pydrake.systems.framework.LeafSystem_TExpressionU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, observed_system: pydrake.systems.framework.System_TExpressionU, observed_system_context: pydrake.systems.framework.Context_TExpressionU, observer_gain) -> None: ...
    def L(self, *args, **kwargs): ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort_TExpressionU: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort_TExpressionU: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort_TExpressionU: ...
    def observer_gain(self, *args, **kwargs): ...

class _TemporaryName_N5drake7systems10estimators18LuenbergerObserverINS_2ad8AutoDiffEEE(pydrake.systems.framework.LeafSystem_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, observed_system: pydrake.systems.framework.System_TAutoDiffXdU, observed_system_context: pydrake.systems.framework.Context_TAutoDiffXdU, observer_gain) -> None: ...
    def L(self, *args, **kwargs): ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort_TAutoDiffXdU: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort_TAutoDiffXdU: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort_TAutoDiffXdU: ...
    def observer_gain(self, *args, **kwargs): ...

class _TemporaryName_N5drake7systems10estimators18LuenbergerObserverINS_8symbolic10ExpressionEEE(pydrake.systems.framework.LeafSystem_TExpressionU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, observed_system: pydrake.systems.framework.System_TExpressionU, observed_system_context: pydrake.systems.framework.Context_TExpressionU, observer_gain) -> None: ...
    def L(self, *args, **kwargs): ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort_TExpressionU: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort_TExpressionU: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort_TExpressionU: ...
    def observer_gain(self, *args, **kwargs): ...



import collections.abc
import enum
import nanobind
import pydrake.geometry
import pydrake.multibody.math
import pydrake.multibody.plant
import pydrake.multibody.tree
import pydrake.solvers
import pydrake.systems.framework
import pydrake.trajectories
from _typeshed import Incomplete
from typing import Callable, ClassVar, overload

class CalcGridPointsOptions:
    max_err: Incomplete
    max_iter: Incomplete
    max_seg_length: Incomplete
    min_points: Incomplete
    def __init__(self, **kwargs) -> None: ...
    def __copy__(self) -> CalcGridPointsOptions: ...
    def __deepcopy__(self, arg: dict) -> CalcGridPointsOptions: ...
    @property
    def __fields__(self): ...

class CentroidalMomentumConstraint(pydrake.solvers.Constraint):
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant_TAutoDiffXdU, model_instances: collections.abc.Sequence[pydrake.multibody.tree.ModelInstanceIndex] | None, plant_context: pydrake.systems.framework.Context_TAutoDiffXdU, angular_only: bool) -> None: ...

class ContactWrench:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def F_Cb_W(self) -> pydrake.multibody.math.SpatialForce: ...
    @property
    def bodyA_index(self) -> pydrake.multibody.tree.BodyIndex: ...
    @property
    def bodyB_index(self) -> pydrake.multibody.tree.BodyIndex: ...
    @property
    def p_WCb_W(self): ...

class ContactWrenchFromForceInWorldFrameEvaluator(pydrake.solvers.EvaluatorBase):
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant_TAutoDiffXdU, context: pydrake.systems.framework.Context_TAutoDiffXdU, geometry_id_pair: tuple[pydrake.geometry.GeometryId, pydrake.geometry.GeometryId]) -> None: ...

class QuaternionEulerIntegrationConstraint(pydrake.solvers.Constraint):
    def __init__(self, allow_quaternion_negation: bool) -> None: ...
    def ComposeVariable(self, *args, **kwargs): ...
    def allow_quaternion_negation(self) -> bool: ...

class SpatialVelocityConstraint(pydrake.solvers.Constraint):
    class AngularVelocityBounds:
        magnitude_lower: float
        magnitude_upper: float
        reference_direction: Incomplete
        theta_bound: float
        def __init__(self) -> None: ...
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant_TAutoDiffXdU, frameA: pydrake.multibody.tree.Frame_TAutoDiffXdU, v_AC_lower, v_AC_upper, frameB: pydrake.multibody.tree.Frame_TAutoDiffXdU, p_BCo, plant_context: pydrake.systems.framework.Context_TAutoDiffXdU, w_AC_bounds: SpatialVelocityConstraint.AngularVelocityBounds | None = ...) -> None: ...

class StaticEquilibriumProblem:
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant_TAutoDiffXdU, context: pydrake.systems.framework.Context_TAutoDiffXdU, ignored_collision_pairs: collections.abc.Set[tuple[pydrake.geometry.GeometryId, pydrake.geometry.GeometryId]]) -> None: ...
    def GetContactWrenchSolution(self, result: pydrake.solvers.MathematicalProgramResult) -> list[ContactWrench]: ...
    def UpdateComplementarityTolerance(self, tol: float) -> None: ...
    def get_mutable_prog(self) -> pydrake.solvers.MathematicalProgram: ...
    def prog(self) -> pydrake.solvers.MathematicalProgram: ...
    def q_vars(self, *args, **kwargs): ...
    def u_vars(self, *args, **kwargs): ...

class Toppra:
    CalcGridPoints: ClassVar[nanobind.nb_func] = ...
    def __init__(self, path: pydrake.trajectories.Trajectory, plant: pydrake.multibody.plant.MultibodyPlant, gridpoints) -> None: ...
    @overload
    def AddFrameAccelerationLimit(self, constraint_frame: pydrake.multibody.tree.Frame, lower_limit, upper_limit, discretization: ToppraDiscretization = ...) -> tuple[pydrake.solvers.BindingTLinearConstraintU, pydrake.solvers.BindingTLinearConstraintU]: ...
    @overload
    def AddFrameAccelerationLimit(self, constraint_frame: pydrake.multibody.tree.Frame, lower_limit: pydrake.trajectories.Trajectory, upper_limit: pydrake.trajectories.Trajectory, discretization: ToppraDiscretization = ...) -> tuple[pydrake.solvers.BindingTLinearConstraintU, pydrake.solvers.BindingTLinearConstraintU]: ...
    @overload
    def AddFrameTranslationalSpeedLimit(self, constraint_frame: pydrake.multibody.tree.Frame, upper_limit: float) -> pydrake.solvers.BindingTBoundingBoxConstraintU: ...
    @overload
    def AddFrameTranslationalSpeedLimit(self, constraint_frame: pydrake.multibody.tree.Frame, upper_limit: pydrake.trajectories.Trajectory) -> pydrake.solvers.BindingTBoundingBoxConstraintU: ...
    def AddFrameVelocityLimit(self, constraint_frame: pydrake.multibody.tree.Frame, lower_limit, upper_limit) -> pydrake.solvers.BindingTBoundingBoxConstraintU: ...
    def AddJointAccelerationLimit(self, lower_limit, upper_limit, discretization: ToppraDiscretization = ...) -> tuple[pydrake.solvers.BindingTLinearConstraintU, pydrake.solvers.BindingTLinearConstraintU]: ...
    def AddJointTorqueLimit(self, lower_limit, upper_limit, discretization: ToppraDiscretization = ...) -> tuple[pydrake.solvers.BindingTLinearConstraintU, pydrake.solvers.BindingTLinearConstraintU]: ...
    def AddJointVelocityLimit(self, lower_limit, upper_limit) -> pydrake.solvers.BindingTBoundingBoxConstraintU: ...
    def SolvePathParameterization(self, s_dot_start: float = ..., s_dot_end: float = ...) -> pydrake.trajectories.PiecewisePolynomial | None: ...
    def constraint_relaxation(self) -> float: ...
    def set_constraint_relaxation(self, relaxation: float) -> None: ...

class ToppraDiscretization(enum.Enum):
    __new__: ClassVar[Callable] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _hashable_values_: ClassVar[list] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[object]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _unhashable_values_map_: ClassVar[dict] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    _value_repr_: ClassVar[None] = ...
    kCollocation: ClassVar[ToppraDiscretization] = ...
    kInterpolation: ClassVar[ToppraDiscretization] = ...
    __nb_enum__: ClassVar[PyCapsule] = ...



# ruff: noqa: F401,F403 (unused-import, import-star)
# ruff: isort: skip_file

import warnings

# Normal symbols.
from .cenic import *  # noqa
from .contact_solvers import *  # noqa
from .inverse_kinematics import *  # noqa
from .fem import *  # noqa
from .math import *  # noqa
from .meshcat import *  # noqa
from .optimization import *  # noqa
from .parsing import *  # noqa
from .plant import *  # noqa
from .rational import *  # noqa
from .tree import *  # noqa

# Submodules.
from .benchmarks.all import *  # noqa

# Main programs.
from . import fix_inertia
from . import mesh_to_model



# ruff: noqa: F403 (import-star)

from .analysis import *
from .controllers import *
from .drawing import *
from .estimators import *
from .framework import *
from .lcm import *
from .perception import *
from .planar_scenegraph_visualizer import *
from .primitives import *
from .pyplot_visualizer import *
from .rendering import *
from .scalar_conversion import *
from .sensors import *
